#![forbid(elided_lifetimes_in_paths, unsafe_code)]

use aoc23::read;
use chumsky::prelude::*;

#[derive(Clone, Copy, Debug, Eq, PartialEq)]
enum Cell {
	Working,
	Broken,
	Unknown
}

impl Cell {
	fn parser() -> impl Parser<char, Self, Error = Simple<char>> {
		choice((
			just('.').map(|_| Self::Working),
			just('#').map(|_| Self::Broken),
			just('?').map(|_| Self::Unknown)
		))
	}
}

struct Pattern {
	grid: Vec<Vec<Cell>>
}

impl Pattern {
	fn parser() -> impl Parser<char, Self, Error = Simple<char>> {
		Cell::parser()
			.repeated()
			.at_least(1)
			.then_ignore(just("\n"))
			.repeated()
			.at_least(1)
			.map(|grid| Self { grid })
	}
}

fn parser() -> impl Parser<char, Vec<Pattern>, Error = Simple<char>> {
	Pattern::parser()
		.separated_by(just("\n"))
		.then_ignore(end())
}

fn find_mirrors_in_patters(patterns: &[Pattern], diff: u32) -> usize {
	let mut rows = Vec::new();
	let mut cols = Vec::new();

	for Pattern { grid } in patterns {
		'i: for i in 1 .. grid.len() {
			let mut d = 0;
			for j in 0 .. i.min(grid.len() - i) {
				assert_eq!(grid[0].len(), grid[i - j - 1].len());
				assert_eq!(grid[0].len(), grid[i + j].len());
				for k in 0 .. grid[0].len() {
					if grid[i - j - 1][k] != grid[i + j][k] {
						d += 1;
						if d > diff {
							continue 'i;
						}
					}
				}
			}
			if d == diff {
				rows.push(i);
			}
		}

		'i: for i in 1 .. grid[0].len() {
			let mut d = 0;
			for j in 0 .. i.min(grid[0].len() - i) {
				for row in grid {
					if row[i - j - 1] != row[i + j] {
						d += 1;
						if d > diff {
							continue 'i;
						}
					}
				}
			}
			if d == diff {
				cols.push(i);
			}
		}
	}

	rows.iter().sum::<usize>() * 100 + cols.iter().sum::<usize>()
}

fn main() -> anyhow::Result<()> {
	let patterns = read("inputs/day13.txt", parser())?;

	println!("{}", find_mirrors_in_patters(&patterns, 0));
	println!("{}", find_mirrors_in_patters(&patterns, 1));

	Ok(())
}
